{%- block doctype -%}
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
  "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
{%- endblock %}
{%- set reldelim1 = reldelim1 is not defined and ' &raquo;' or reldelim1 %}
{%- set reldelim2 = reldelim2 is not defined and ' |' or reldelim2 %}
{%- macro relbar() %}
    <div class="related">
      <h3>Navigation</h3>
      <ul>
        {%- for rellink in rellinks %}
        <li class="right" {% if loop.first %}style="margin-right: 10px"{% endif %}>
          <a href="{{ pathto(rellink[0]) }}" title="{{ rellink[1]|striptags }}"
             accesskey="{{ rellink[2] }}">{{ rellink[3] }}</a>
          {%- if not loop.first %}{{ reldelim2 }}{% endif %}</li>
        {%- endfor %}
        {%- block rootrellink %}
        <li><a href="{{ pathto('index') }}">{{ shorttitle }}</a>{{ reldelim1 }}</li>
        {%- endblock %}
        {%- for parent in parents %}
          <li><a href="{{ parent.link|e }}" accesskey="U">{{ parent.title }}</a>{{ reldelim1 }}</li>
        {%- endfor %}
        {%- block relbaritems %}{% endblock %}
      </ul>
    </div>
{%- endmacro %}
{%- macro sidebar() %}
      {%- if builder != 'htmlhelp' %}
      <div class="sphinxsidebar">
        <div class="sphinxsidebarwrapper">
          {%- block sidebarlogo %}
          {%- if logo %}
            <p class="logo"><img class="logo" src="{{ pathto('_static/' + logo, 1) }}" alt="Logo"/></p>
          {%- endif %}
          {%- endblock %}
          {%- block sidebartoc %}
          {%- if display_toc %}
            <h3>Table Of Contents</h3>
            {{ toc }}
          {%- endif %}
          {%- endblock %}
          {%- block sidebarrel %}
          {%- if prev %}
            <h4>Previous topic</h4>
            <p class="topless"><a href="{{ prev.link|e }}" title="previous chapter">{{ prev.title }}</a></p>
          {%- endif %}
          {%- if next %}
            <h4>Next topic</h4>
            <p class="topless"><a href="{{ next.link|e }}" title="next chapter">{{ next.title }}</a></p>
          {%- endif %}
          {%- endblock %}
          {%- if sourcename %}
            <!--<h3>This Page</h3>
            <ul class="this-page-menu">
            {%- if builder == 'web' %}
              <li><a href="#comments">Comments ({{ comments|length }} so far)</a></li>
              <li><a href="{{ pathto('@edit/' + sourcename)|e }}">Suggest Change</a></li>
              <li><a href="{{ pathto('@source/' + sourcename)|e }}">Show Source</a></li>
            {%- elif builder == 'html' %}
              <li><a href="{{ pathto('_sources/' + sourcename, true)|e }}">Show Source</a></li>
            {%- endif %}
            </ul>-->
          {%- endif %}
          {%- if customsidebar %}
          {{ rendertemplate(customsidebar) }}
          {%- endif %}
          {%- block sidebarsearch %}
          {%- if pagename != "search" %}
            <h3>{{ builder == 'web' and 'Keyword' or 'Quick' }} search</h3>
            <form class="search" action="{{ pathto('search') }}" method="get">
              <input type="text" name="q" size="18" /> <input type="submit" value="Go" />
              <input type="hidden" name="check_keywords" value="yes" />
              <input type="hidden" name="area" value="default" />
            </form>
            {%- if builder == 'web' %}
            <p style="font-size: 90%">Enter a module, class or function name.</p>
            {%- endif %}
          {%- endif %}
          {%- endblock %}
        </div>
      </div>
      {%- endif %}
{%- endmacro -%}

<html xmlns="http://www.w3.org/1999/xhtml">
  <head>
    <meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
    {%- if builder != 'htmlhelp' %}
      {%- set titlesuffix = " &mdash; " + docstitle %}
    {%- endif %}
    <title>{{ title|striptags }}{{ titlesuffix }}</title>
    {%- if builder == 'web' %}
    <link rel="stylesheet" href="{{ pathto('index') }}?do=stylesheet{%
      if in_admin_panel %}&admin=yes{% endif %}" type="text/css" />
    {%- for link, type, title in page_links %}
    <link rel="alternate" type="{{ type|e(true) }}" title="{{ title|e(true) }}" href="{{ link|e(true) }}" />
    {%- endfor %}
    {%- else %}
    <link rel="stylesheet" href="{{ pathto('_static/' + style, 1) }}" type="text/css" />
    <link rel="stylesheet" href="{{ pathto('_static/pygments.css', 1) }}" type="text/css" />
    {%- endif %}
    {%- if builder != 'htmlhelp' %}
    <script type="text/javascript">
      var DOCUMENTATION_OPTIONS = {
          URL_ROOT:    '{{ pathto("", 1) }}',
          VERSION:     '{{ release }}',
          COLLAPSE_MODINDEX: false,
          FILE_SUFFIX: '{{ file_suffix }}'
      };
    </script>
    <script type="text/javascript" src="{{ pathto('_static/jquery.js', 1) }}"></script>
    <script type="text/javascript" src="{{ pathto('_static/interface.js', 1) }}"></script>
    <script type="text/javascript" src="{{ pathto('_static/doctools.js', 1) }}"></script>
    <script type="text/javascript" src="{{ pathto('_static/searchtools.js', 1) }}"></script>
    {%- if use_opensearch %}
    <link rel="search" type="application/opensearchdescription+xml"
          title="Search within {{ docstitle }}"
          href="{{ pathto('_static/opensearch.xml', 1) }}"/>
    {%- endif %}
    {%- if favicon %}
    <link rel="shortcut icon" href="{{ pathto('_static/' + favicon, 1) }}"/>
    {%- endif %}
    {%- endif %}
{%- block rellinks %}
    {%- if hasdoc('about') %}
    <link rel="author" title="About these documents" href="{{ pathto('about') }}" />
    {%- endif %}
    <link rel="contents" title="Global table of contents" href="{{ pathto('contents') }}" />
    <link rel="index" title="Global index" href="{{ pathto('genindex') }}" />
    <link rel="search" title="Search" href="{{ pathto('search') }}" />
    {%- if hasdoc('copyright') %}
    <link rel="copyright" title="Copyright" href="{{ pathto('copyright') }}" />
    {%- endif %}
    <link rel="top" title="{{ docstitle }}" href="{{ pathto('index') }}" />
    {%- if parents %}
    <link rel="up" title="{{ parents[-1].title|striptags }}" href="{{ parents[-1].link|e }}" />
    {%- endif %}
    {%- if next %}
    <link rel="next" title="{{ next.title|striptags }}" href="{{ next.link|e }}" />
    {%- endif %}
    {%- if prev %}
    <link rel="prev" title="{{ prev.title|striptags }}" href="{{ prev.link|e }}" />
    {%- endif %}
{%- endblock %}
{%- block extrahead %}{% endblock %}
  </head>
  <body>

{% block logilablogo %}
<div class="logilablogo">
	<a class="logogo" href="http://www.cubicweb.org"><img border="0" src="{{ pathto('_static/cubicweb.png', 1) }}"/></a>
  </div>
{% endblock %}

{%- block relbar1 %}{{ relbar() }}{% endblock %}

{%- block sidebar1 %}{# possible location for sidebar #}{% endblock %}

{%- block document %}
    <div class="document">
      <div class="documentwrapper">
      {%- if builder != 'htmlhelp' %}
        <div class="bodywrapper">
      {%- endif %}
          <div class="body">
            {% block body %}{% endblock %}
          </div>
      {%- if builder != 'htmlhelp' %}
        </div>
      {%- endif %}
      </div>
{%- endblock %}

{%- block sidebar2 %}{{ sidebar() }}{% endblock %}
      <div class="clearer"></div>
    </div>

{%- block relbar2 %}{{ relbar() }}{% endblock %}

{%- block footer %}
    <div class="footer">
    {%- if hasdoc('copyright') %}
      &copy; <a href="{{ pathto('copyright') }}">Copyright</a> {{ copyright }}.
    {%- else %}
      &copy; Copyright {{ copyright }}.
    {%- endif %}
    {%- if last_updated %}
      Last updated on {{ last_updated }}.
    {%- endif %}
    {%- if show_sphinx %}
      Created using <a href="http://sphinx.pocoo.org/">Sphinx</a>.
    {%- endif %}
    </div>
{%- endblock %}
  </body>
</html>
