/***************************************/
/* filter table form                   */
/***************************************/

table.filter th {
  font-weight: bold;
  background: #ebe8d9 url("button.png") repeat-x;
  padding: 0.3em;
  border-bottom: 1px solid %(incontextBoxBodyBgColor)s;
  text-align: left;
}

table.filter div.facet {
  padding: 0.6em 0.2em;
  margin: 0em 1em;
  border: 1px solid #ccc;
}

table.filter div.facetTitle {
  font-weight: bold;
}

div.tableActionsBox {
}

div.tableActionsBox .popup {
  border-radius: 5px;
  background: %(incontextBoxBodyBgColor)s;
  box-shadow: 3px 3px 3px Grey;
}

div.tableActionsBox li {
  background: none;
  /* we should probably get rid of ul/li structure because
     of the spurious space consumed by the bullet */
  margin-right: .3em;
  margin-left: -.3em;
}

