sync_schema_props_perms('state_of')
sync_schema_props_perms('transition_of')
sync_schema_props_perms('State')
sync_schema_props_perms('BaseTransition')
sync_schema_props_perms('Transition')
sync_schema_props_perms('WorkflowTransition')
