/* styles for preferences form (views/management.py)
 *
 *  :organization: Logilab
 *  :copyright: 2003-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
 *  :contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
 */

div.propertiesform {
  margin-bottom: 1.2857em;
  line-height: 1.2857em;
  font-size: %(h3FontSize)s;
}

div.propertiesform a {
  display: block;
  margin: 10px 0px 6px 0px;
  padding-left: 16px;
  font-weight: bold;
  color: #000;
  background: transparent url("puce_down.png") 3px center no-repeat;
  text-decoration:none;
}

div.propertiesform a:hover {
  background-color: %(listingHeaderBgColor)s;
}

.preferences .validateButton {
  margin-top: 0px;
 }

fieldset.preferences {
  margin: 7px 1em 0;
  padding: 2px 6px 6px;
  border : 1px solid %(pageContentBorderColor)s;
}

div.component {
  margin: 0 0 1em 16px;
}

a.componentTitle {
 font-weight: bold;
 color: #000
}

a.componentTitle:visited {
  color: #000;
}

div.preffield {
  margin-bottom: 5px;
  padding: 2px 5px;
  background: %(listingHeaderBgColor)s;
}

div.prefinput {
  margin: .3em;
}

div.prefinput select.changed,
div.prefinput input.changed {
  border: 1px solid #000;
  font-weight: bold;
}

.prefinput input.error {
  border:1px solid red !important;
  color:red;
  padding-right:1em;
}

div.formsg {
  font-weight: bold;
  margin: 0.5em 0px;
}

div.critical {
  color: red;
  padding-left: 20px;
  background: #fff url(critical.png) no-repeat;
 }

div.formsg .msg {
  color: green;
}

.helper {
  font-size: 96%;
  color: %(helperColor)s;
  padding: 0;
}

div.prefinput .helper:hover {
  color: #000;
  cursor: default;
}

div.openlink {
  display: inline;
 }
