/* styles for the login popup and login form
 *
 *  :organization: Logilab
 *  :copyright: 2003-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
 *  :contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
 */

div.popupLoginBox {
  position: absolute;
  z-index: 400;
  right: 0px;
  width: 26em;
  padding: 0px 1px 1px;
  background: %(listingBorderColor)s;
}

div.popupLoginBox label{
  font-weight: bold;
}

div.popupLoginBox div.loginContent {
  background: #e6e4ce;
  padding: 5px 3px 4px;
}

div#loginBox {
  position : absolute;
  top: 15%;
  left : 50%;
  margin-left: -14em;
  width: 28em;
  background: #fff;
  border: 2px solid %(incontextBoxBodyBgColor)s;
  padding-bottom: 0.5em;
  text-align: center;
}

div#loginBox h1 {
  color: %(aColor)s;
  font-size: 140%;
}

div.loginTitle {
  color: #fff;
  font-weight: bold;
  font-size: 140%;
  text-align: center;
  padding: 3px 0px;
  background: %(headerBgColor)s url("banner.png") repeat-x top left;
}

div#loginBox div.loginContent form {
  padding-top: 1em;
  width: 90%;
  margin: auto;
}


.popupLoginBox table td {
  padding: 0px 3px;
  white-space: nowrap;
}

.loginContent table {
  padding: 0px 0.5em;
  margin: auto;
}

#loginBox table td {
  padding: 0px 3px 0.6em;
  white-space: nowrap;
}

#loginBox .loginButton {
  margin-top: 0.6em;
 }

.loginContent input.data {
  width: 12em;
}

/* This is seriously debatable
   These buttons render much more clearly as standard/default buttons
 */
.loginButton {
  border: 1px solid #edecd2;
  border-color: #edecd2 %(incontextBoxBodyBgColor)s %(incontextBoxBodyBgColor)s  #edecd2;
  margin: 2px 0px 0px;
  background: #f0eff0 url("gradient-grey-up.png") left top repeat-x;
}

.loginContent .formButtonBar {
  float: right;
}
