/*The included Entypo font have been created by Daniel Bruce (www.entypo.com) as 
proposed by fontello (https://github.com/fontello/entypo). 
The Entypo pictograms are licensed under CC BY 3.0 and the font under 
SIL Open Font License.*/
@font-face {
  font-family: 'entypo';
  src: url('entypo.eot?8644018');
  src: url('entypo.eot?8644018#iefix') format('embedded-opentype'),
       url('entypo.woff?8644018') format('woff'),
       url('entypo.ttf?8644018') format('truetype'),
       url('entypo.svg?8644018#entypo') format('svg');
  font-weight: normal;
  font-style: normal;
}
/* Chrome hack: SVG is rendered more smooth in Windozze. 100% magic, uncomment if you need it. */
/* Note, that will break hinting! In other OS-es font will be not as sharp as it could be */
/*
@media screen and (-webkit-min-device-pixel-ratio:0) {
  @font-face {
    font-family: 'entypo';
    src: url('../font/entypo.svg?8644018#entypo') format('svg');
  }
}
*/

[class^="icon-"]:before,
[class*=" icon-"]:before {
  font-family: "entypo";
  font-style: normal;
  font-weight: normal;
  speak: none;

  display: inline-block;
  text-decoration: inherit;
  width: 1em;
  margin-right: .1em;
  text-align: center;
  /* opacity: .8; */

  /* For safety - reset parent styles, that can break glyph codes*/
  font-variant: normal;
  text-transform: none;

  /* fix buttons height, for twitter bootstrap */
  line-height: 1em;

  /* Animation center compensation - magrins should be symmetric */
  /* remove if not needed */
  margin-left: .1em;

  /* you can be more comfortable with increased icons size */
  font-size: 160%;
  vertical-align: middle;

  /* Uncomment for 3D effect */
  /* text-shadow: 1px 1px 1px rgba(127, 127, 127, 0.3); */
}

.icon-note:before { content: '\266a'; } /* '♪' */
.icon-note-beamed:before { content: '\266b'; } /* '♫' */
.icon-music:before { content: '\1f3b5'; } /* '\1f3b5' */
.icon-search:before { content: '\1f50d'; } /* '\1f50d' */
.icon-flashlight:before { content: '\1f526'; } /* '\1f526' */
.icon-mail:before { content: '\2709'; } /* '✉' */
.icon-heart:before { content: '\2665'; } /* '♥' */
.icon-heart-empty:before { content: '\2661'; } /* '♡' */
.icon-star:before { content: '\2605'; } /* '★' */
.icon-star-empty:before { content: '\2606'; } /* '☆' */
.icon-user:before { content: '\1f464'; }
.icon-users:before { content: '\1f465'; }
.icon-user-add:before { content: '\e700'; } /* '' */
.icon-video:before { content: '\1f3ac'; }
.icon-picture:before { content: '\1f304'; }
.icon-camera:before { content: '\1f4f7'; }
.icon-layout:before { content: '\268f'; } /* '⚏' */
.icon-menu:before { content: '\2630'; } /* '☰' */
.icon-check:before { content: '\2713'; } /* '✓' */
.icon-cancel:before { content: '\2715'; } /* '✕' */
.icon-cancel-circled:before { content: '\2716'; } /* '✖' */
.icon-cancel-squared:before { content: '\274e'; } /* '❎' */
.icon-plus:before { content: '\2b'; } /* '+' */
.icon-plus-circled:before { content: '\2795'; } /* '➕' */
.icon-plus-squared:before { content: '\229e'; } /* '⊞' */
.icon-minus:before { content: '\2d'; } /* '-' */
.icon-minus-circled:before { content: '\2796'; } /* '➖' */
.icon-minus-squared:before { content: '\229f'; } /* '⊟' */
.icon-help:before { content: '\2753'; } /* '❓' */
.icon-help-circled:before { content: '\e704'; } /* '' */
.icon-info:before { content: '\2139'; } /* 'ℹ' */
.icon-info-circled:before { content: '\e705'; } /* '' */
.icon-back:before { content: '\1f519'; }
.icon-home:before { content: '\2302'; } /* '⌂' */
.icon-link:before { content: '\1f517'; }
.icon-attach:before { content: '\1f4ce'; }
.icon-lock:before { content: '\1f512'; }
.icon-lock-open:before { content: '\1f513'; }
.icon-eye:before { content: '\e70a'; } /* '' */
.icon-tag:before { content: '\e70c'; } /* '' */
.icon-bookmark:before { content: '\1f516'; }
.icon-bookmarks:before { content: '\1f4d1'; }
.icon-flag:before { content: '\2691'; } /* '⚑' */
.icon-thumbs-up:before { content: '\1f44d'; }
.icon-thumbs-down:before { content: '\1f44e'; }
.icon-download:before { content: '\1f4e5'; }
.icon-upload:before { content: '\1f4e4'; }
.icon-upload-cloud:before { content: '\e711'; } /* '' */
.icon-reply:before { content: '\e712'; } /* '' */
.icon-reply-all:before { content: '\e713'; } /* '' */
.icon-forward:before { content: '\27a6'; } /* '➦' */
.icon-quote:before { content: '\275e'; } /* '❞' */
.icon-code:before { content: '\e714'; } /* '' */
.icon-export:before { content: '\e715'; } /* '' */
.icon-pencil:before { content: '\270e'; } /* '✎' */
.icon-feather:before { content: '\2712'; } /* '✒' */
.icon-print:before { content: '\e716'; } /* '' */
.icon-retweet:before { content: '\e717'; } /* '' */
.icon-keyboard:before { content: '\2328'; } /* '⌨' */
.icon-comment:before { content: '\e718'; } /* '' */
.icon-chat:before { content: '\e720'; } /* '' */
.icon-bell:before { content: '\1f514'; }
.icon-attention:before { content: '\26a0'; } /* '⚠' */
.icon-alert:before { content: '\1f4a5'; }
.icon-vcard:before { content: '\e722'; } /* '' */
.icon-address:before { content: '\e723'; } /* '' */
.icon-location:before { content: '\e724'; } /* '' */
.icon-map:before { content: '\e727'; } /* '' */
.icon-direction:before { content: '\27a2'; } /* '➢' */
.icon-compass:before { content: '\e728'; } /* '' */
.icon-cup:before { content: '\2615'; } /* '☕' */
.icon-trash:before { content: '\e729'; } /* '' */
.icon-doc:before { content: '\e730'; } /* '' */
.icon-docs:before { content: '\e736'; } /* '' */
.icon-doc-landscape:before { content: '\e737'; } /* '' */
.icon-doc-text:before { content: '\1f4c4'; }
.icon-doc-text-inv:before { content: '\e731'; } /* '' */
.icon-newspaper:before { content: '\1f4f0'; }
.icon-book-open:before { content: '\1f4d6'; }
.icon-book:before { content: '\1f4d5'; }
.icon-folder:before { content: '\1f4c1'; }
.icon-archive:before { content: '\e738'; } /* '' */
.icon-box:before { content: '\1f4e6'; }
.icon-rss:before { content: '\e73a'; } /* '' */
.icon-phone:before { content: '\1f4de'; }
.icon-cog:before { content: '\2699'; } /* '⚙' */
.icon-tools:before { content: '\2692'; } /* '⚒' */
.icon-share:before { content: '\e73c'; } /* '' */
.icon-shareable:before { content: '\e73e'; } /* '' */
.icon-basket:before { content: '\e73d'; } /* '' */
.icon-bag:before { content: '\1f45c'; }
.icon-calendar:before { content: '\1f4c5'; }
.icon-login:before { content: '\e740'; } /* '' */
.icon-logout:before { content: '\e741'; } /* '' */
.icon-mic:before { content: '\1f3a4'; }
.icon-mute:before { content: '\1f507'; }
.icon-sound:before { content: '\1f50a'; }
.icon-volume:before { content: '\e742'; } /* '' */
.icon-clock:before { content: '\1f554'; }
.icon-hourglass:before { content: '\23f3'; } /* '⏳' */
.icon-lamp:before { content: '\1f4a1'; }
.icon-light-down:before { content: '\1f505'; }
.icon-light-up:before { content: '\1f506'; }
.icon-adjust:before { content: '\25d1'; } /* '◑' */
.icon-block:before { content: '\1f6ab'; }
.icon-resize-full:before { content: '\e744'; } /* '' */
.icon-resize-small:before { content: '\e746'; } /* '' */
.icon-popup:before { content: '\e74c'; } /* '' */
.icon-publish:before { content: '\e74d'; } /* '' */
.icon-window:before { content: '\e74e'; } /* '' */
.icon-arrow-combo:before { content: '\e74f'; } /* '' */
.icon-down-circled:before { content: '\e758'; } /* '' */
.icon-left-circled:before { content: '\e759'; } /* '' */
.icon-right-circled:before { content: '\e75a'; } /* '' */
.icon-up-circled:before { content: '\e75b'; } /* '' */
.icon-down-open:before { content: '\e75c'; } /* '' */
.icon-left-open:before { content: '\e75d'; } /* '' */
.icon-right-open:before { content: '\e75e'; } /* '' */
.icon-up-open:before { content: '\e75f'; } /* '' */
.icon-down-open-mini:before { content: '\e760'; } /* '' */
.icon-left-open-mini:before { content: '\e761'; } /* '' */
.icon-right-open-mini:before { content: '\e762'; } /* '' */
.icon-up-open-mini:before { content: '\e763'; } /* '' */
.icon-down-open-big:before { content: '\e764'; } /* '' */
.icon-left-open-big:before { content: '\e765'; } /* '' */
.icon-right-open-big:before { content: '\e766'; } /* '' */
.icon-up-open-big:before { content: '\e767'; } /* '' */
.icon-down:before { content: '\2b07'; } /* '⬇' */
.icon-left:before { content: '\2b05'; } /* '⬅' */
.icon-right:before { content: '\27a1'; } /* '➡' */
.icon-up:before { content: '\2b06'; } /* '⬆' */
.icon-down-dir:before { content: '\25be'; } /* '▾' */
.icon-left-dir:before { content: '\25c2'; } /* '◂' */
.icon-right-dir:before { content: '\25b8'; } /* '▸' */
.icon-up-dir:before { content: '\25b4'; } /* '▴' */
.icon-down-bold:before { content: '\e4b0'; } /* '' */
.icon-left-bold:before { content: '\e4ad'; } /* '' */
.icon-right-bold:before { content: '\e4ae'; } /* '' */
.icon-up-bold:before { content: '\e4af'; } /* '' */
.icon-down-thin:before { content: '\2193'; } /* '↓' */
.icon-left-thin:before { content: '\2190'; } /* '←' */
.icon-right-thin:before { content: '\2192'; } /* '→' */
.icon-up-thin:before { content: '\2191'; } /* '↑' */
.icon-ccw:before { content: '\27f2'; } /* '⟲' */
.icon-cw:before { content: '\27f3'; } /* '⟳' */
.icon-arrows-ccw:before { content: '\1f504'; }
.icon-level-down:before { content: '\21b3'; } /* '↳' */
.icon-level-up:before { content: '\21b0'; } /* '↰' */
.icon-shuffle:before { content: '\1f500'; }
.icon-loop:before { content: '\1f501'; }
.icon-switch:before { content: '\21c6'; } /* '⇆' */
.icon-play:before { content: '\25b6'; } /* '▶' */
.icon-stop:before { content: '\25a0'; } /* '■' */
.icon-pause:before { content: '\2389'; } /* '⎉' */
.icon-record:before { content: '\26ab'; } /* '⚫' */
.icon-to-end:before { content: '\23ed'; } /* '⏭' */
.icon-to-start:before { content: '\23ee'; } /* '⏮' */
.icon-fast-forward:before { content: '\23e9'; } /* '⏩' */
.icon-fast-backward:before { content: '\23ea'; } /* '⏪' */
.icon-progress-0:before { content: '\e768'; } /* '' */
.icon-progress-1:before { content: '\e769'; } /* '' */
.icon-progress-2:before { content: '\e76a'; } /* '' */
.icon-progress-3:before { content: '\e76b'; } /* '' */
.icon-target:before { content: '\1f3af'; }
.icon-palette:before { content: '\1f3a8'; }
.icon-list:before { content: '\e005'; } /* '' */
.icon-list-add:before { content: '\e003'; } /* '' */
.icon-signal:before { content: '\1f4f6'; }
.icon-trophy:before { content: '\1f3c6'; }
.icon-battery:before { content: '\1f50b'; }
.icon-back-in-time:before { content: '\e771'; } /* '' */
.icon-monitor:before { content: '\1f4bb'; }
.icon-mobile:before { content: '\1f4f1'; }
.icon-network:before { content: '\e776'; } /* '' */
.icon-cd:before { content: '\1f4bf'; }
.icon-inbox:before { content: '\e777'; } /* '' */
.icon-install:before { content: '\e778'; } /* '' */
.icon-globe:before { content: '\1f30e'; }
.icon-cloud:before { content: '\2601'; } /* '☁' */
.icon-cloud-thunder:before { content: '\26c8'; } /* '⛈' */
.icon-flash:before { content: '\26a1'; } /* '⚡' */
.icon-moon:before { content: '\263d'; } /* '☽' */
.icon-flight:before { content: '\2708'; } /* '✈' */
.icon-paper-plane:before { content: '\e79b'; } /* '' */
.icon-leaf:before { content: '\1f342'; }
.icon-lifebuoy:before { content: '\e788'; } /* '' */
.icon-mouse:before { content: '\e789'; } /* '' */
.icon-briefcase:before { content: '\1f4bc'; }
.icon-suitcase:before { content: '\e78e'; } /* '' */
.icon-dot:before { content: '\e78b'; } /* '' */
.icon-dot-2:before { content: '\e78c'; } /* '' */
.icon-dot-3:before { content: '\e78d'; } /* '' */
.icon-brush:before { content: '\e79a'; } /* '' */
.icon-magnet:before { content: '\e7a1'; } /* '' */
.icon-infinity:before { content: '\221e'; } /* '∞' */
.icon-erase:before { content: '\232b'; } /* '⌫' */
.icon-chart-pie:before { content: '\e751'; } /* '' */
.icon-chart-line:before { content: '\1f4c8'; }
.icon-chart-bar:before { content: '\1f4ca'; }
.icon-chart-area:before { content: '\1f53e'; }
.icon-tape:before { content: '\2707'; } /* '✇' */
.icon-graduation-cap:before { content: '\1f393'; }
.icon-language:before { content: '\e752'; } /* '' */
.icon-ticket:before { content: '\1f3ab'; }
.icon-water:before { content: '\1f4a6'; }
.icon-droplet:before { content: '\1f4a7'; }
.icon-air:before { content: '\e753'; } /* '' */
.icon-credit-card:before { content: '\1f4b3'; }
.icon-floppy:before { content: '\1f4be'; }
.icon-clipboard:before { content: '\1f4cb'; }
.icon-megaphone:before { content: '\1f4e3'; }
.icon-database:before { content: '\e754'; } /* '' */
.icon-drive:before { content: '\e755'; } /* '' */
.icon-bucket:before { content: '\e756'; } /* '' */
.icon-thermometer:before { content: '\e757'; } /* '' */
.icon-key:before { content: '\1f511'; }
.icon-flow-cascade:before { content: '\e790'; } /* '' */
.icon-flow-branch:before { content: '\e791'; } /* '' */
.icon-flow-tree:before { content: '\e792'; } /* '' */
.icon-flow-line:before { content: '\e793'; } /* '' */
.icon-flow-parallel:before { content: '\e794'; } /* '' */
.icon-rocket:before { content: '\1f680'; }
.icon-gauge:before { content: '\e7a2'; } /* '' */
.icon-traffic-cone:before { content: '\e7a3'; } /* '' */
.icon-cc:before { content: '\e7a5'; } /* '' */
.icon-cc-by:before { content: '\e7a6'; } /* '' */
.icon-cc-nc:before { content: '\e7a7'; } /* '' */
.icon-cc-nc-eu:before { content: '\e7a8'; } /* '' */
.icon-cc-nc-jp:before { content: '\e7a9'; } /* '' */
.icon-cc-sa:before { content: '\e7aa'; } /* '' */
.icon-cc-nd:before { content: '\e7ab'; } /* '' */
.icon-cc-pd:before { content: '\e7ac'; } /* '' */
.icon-cc-zero:before { content: '\e7ad'; } /* '' */
.icon-cc-share:before { content: '\e7ae'; } /* '' */
.icon-cc-remix:before { content: '\e7af'; } /* '' */
.icon-github:before { content: '\f300'; } /* '' */
.icon-github-circled:before { content: '\f301'; } /* '' */
.icon-flickr:before { content: '\f303'; } /* '' */
.icon-flickr-circled:before { content: '\f304'; } /* '' */
.icon-vimeo:before { content: '\f306'; } /* '' */
.icon-vimeo-circled:before { content: '\f307'; } /* '' */
.icon-twitter:before { content: '\f309'; } /* '' */
.icon-twitter-circled:before { content: '\f30a'; } /* '' */
.icon-facebook:before { content: '\f30c'; } /* '' */
.icon-facebook-circled:before { content: '\f30d'; } /* '' */
.icon-facebook-squared:before { content: '\f30e'; } /* '' */
.icon-gplus:before { content: '\f30f'; } /* '' */
.icon-gplus-circled:before { content: '\f310'; } /* '' */
.icon-pinterest:before { content: '\f312'; } /* '' */
.icon-pinterest-circled:before { content: '\f313'; } /* '' */
.icon-tumblr:before { content: '\f315'; } /* '' */
.icon-tumblr-circled:before { content: '\f316'; } /* '' */
.icon-linkedin:before { content: '\f318'; } /* '' */
.icon-linkedin-circled:before { content: '\f319'; } /* '' */
.icon-dribbble:before { content: '\f31b'; } /* '' */
.icon-dribbble-circled:before { content: '\f31c'; } /* '' */
.icon-stumbleupon:before { content: '\f31e'; } /* '' */
.icon-stumbleupon-circled:before { content: '\f31f'; } /* '' */
.icon-lastfm:before { content: '\f321'; } /* '' */
.icon-lastfm-circled:before { content: '\f322'; } /* '' */
.icon-rdio:before { content: '\f324'; } /* '' */
.icon-rdio-circled:before { content: '\f325'; } /* '' */
.icon-spotify:before { content: '\f327'; } /* '' */
.icon-spotify-circled:before { content: '\f328'; } /* '' */
.icon-qq:before { content: '\f32a'; } /* '' */
.icon-instagram:before { content: '\f32d'; } /* '' */
.icon-dropbox:before { content: '\f330'; } /* '' */
.icon-evernote:before { content: '\f333'; } /* '' */
.icon-flattr:before { content: '\f336'; } /* '' */
.icon-skype:before { content: '\f339'; } /* '' */
.icon-skype-circled:before { content: '\f33a'; } /* '' */
.icon-renren:before { content: '\f33c'; } /* '' */
.icon-sina-weibo:before { content: '\f33f'; } /* '' */
.icon-paypal:before { content: '\f342'; } /* '' */
.icon-picasa:before { content: '\f345'; } /* '' */
.icon-soundcloud:before { content: '\f348'; } /* '' */
.icon-mixi:before { content: '\f34b'; } /* '' */
.icon-behance:before { content: '\f34e'; } /* '' */
.icon-google-circles:before { content: '\f351'; } /* '' */
.icon-vkontakte:before { content: '\f354'; } /* '' */
.icon-smashing:before { content: '\f357'; } /* '' */
.icon-sweden:before { content: '\f601'; } /* '' */
.icon-db-shape:before { content: '\f600'; } /* '' */
.icon-logo-db:before { content: '\f603'; } /* '' */

